@@ -1,20 +1,23 @@
-use super::{Queryable, Value, ValueConvertError};
+use super::{Error, Queryable, Value};
 use std::convert::TryFrom;
 use std::str::FromStr;
 use uuid::Uuid;
 
 impl<'a> Queryable<'a> for Uuid {
+    fn name(&self) -> &'static str {
+        "Uuid"
+    }
     fn data(&self) -> Option<Value> {
         Some(Value::String(format!("{}", self)))
     }
 }
 
 impl TryFrom<Value> for Uuid {
-    type Error = ValueConvertError;
-    fn try_from(value: Value) -> Result<Self, ValueConvertError> {
+    type Error = Error;
+    fn try_from(value: Value) -> Result<Self, Self::Error> {
         match value {
-            Value::String(s) => Uuid::from_str(&s).map_err(|_| ValueConvertError),
-            _ => Err(ValueConvertError),
+            Value::String(s) => Uuid::from_str(&s).map_err(|_| Error::TypeError),
+            _ => Err(Error::TypeError),
         }
     }
 }
